package jp.riken.plot2d.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import javax.swing.*;
import jp.riken.common.api.*;
import jp.riken.common.exception.*;
import jp.riken.plot2d.data.*;
import jp.riken.plot2d.renderer.*;

public class Plot2dFrame extends JFrame implements ActionListener,
        GuiController, Plot2d {

    private static final long serialVersionUID = 1L;

    private static final String TITLE = "";
    private static final int DEFAULT_WIDTH  = 615;
    private static final int DEFAULT_HEIGHT = 475;
    private static final int PANEL_WIDTH  = 600;
    private static final int PANEL_HEIGHT = 400;

    private Viewer viewer;

    private ImageRenderer renderer;
    private RenderingContext context;
    private PlotData data;
    private MousePickHandler mousePickHandler;
    private YaxisRangeDialog yaxisRangeDialog;

    private int plotWidth;  // 0 when scale with frame
    private int plotHeight; // 0 when scale with frame

    private boolean standalone;
    private File currDir;

    // components
    private PlotPanel plotPanel;
    private JScrollPane scrollPane;

    // menu File
    private JMenuItem miFileOpenDataXVG;
    private JMenuItem miFileOpenDataXPM;
    private JMenuItem miFileOpenDataPCA;
    private JMenuItem miFileExit;
    
    // menu X-axis
    private JRadioButtonMenuItem miXaxisFrame;
    private JRadioButtonMenuItem miXaxis100;
    private JRadioButtonMenuItem miXaxis200;
    private JRadioButtonMenuItem miXaxis400;
    private JRadioButtonMenuItem miXaxis800;
    private JRadioButtonMenuItem miXaxis1600;
    
    // menu Y-axis
    private JRadioButtonMenuItem miYaxisFrame;
    private JRadioButtonMenuItem miYaxis100;
    private JRadioButtonMenuItem miYaxis200;
    private JRadioButtonMenuItem miYaxis400;
    private JRadioButtonMenuItem miYaxis800;
    private JRadioButtonMenuItem miYaxis1600;
    private JMenuItem miYaxisRange;

    public Plot2dFrame(boolean standalone) {
        // Look & Feel
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            SwingUtilities.updateComponentTreeUI(this);
        } catch (Exception e) {
            e.printStackTrace();
        }

        this.standalone = standalone;

        // WindowListner
        addWindowListener(new WindowAdapter() {
            public void windowOpened(WindowEvent e) {
                init();
            }
            public void windowClosing(WindowEvent e) {
                exit();
            }
        });

        setSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);
        setTitle(TITLE);

        context = new RenderingContext();
        context.setRatioImageInitX(0.15);
        context.setRatioImageTermX(0.84);
        context.setRatioImageInitY(0.15);
        context.setRatioImageTermY(0.90);
        mousePickHandler = new MousePickHandler(context);

        createMenu();
        createComponents();

        if (standalone) {
            currDir = new File(".");
        }
    }

    private void init() {
    }

    private void exit() {
        if (viewer != null) {
            viewer.removePlot2d(this);
        }
        dispose();
        if (standalone) {
            System.exit(0);
        }
    }

    @Override
    public void dispose() {
        if (yaxisRangeDialog != null) {
            yaxisRangeDialog.dispose();
        }
        super.dispose();
    }

    public void setCenterTo(Component owner) {
        Point ownerLocation = owner.getLocation();
        Dimension ownerSize = owner.getSize();
        Dimension size = getSize();
        setLocation(
                ownerLocation.x + (ownerSize.width - size.width) / 2,
                ownerLocation.y + (ownerSize.height - size.height) / 2);
    }

    public PlotData getData() {
        return data;
    }

    //--------------------------------------------------------------------
    // Implements Plot2d
    //--------------------------------------------------------------------
    public void setViewer(Viewer viewer) {
        this.viewer = viewer;
    }

    public void setTime(float time) {
        if (data.isTimeAware()) {
            data.setTime(time);
            refreshImage();
        }
    }

    //--------------------------------------------------------------------
    // Implements GuiController
    //--------------------------------------------------------------------
    public void mousePicked(int x, int y) {
        if (data != null && data.isPickAware()) {
            Float value = mousePickHandler.picked(x, y, data);
            if (value != null) {
                refreshImage();
                if (viewer != null && data.isXaxisTime()) {
                    viewer.setTime(value * data.getXaxisScale());
                }
            }
        }
    }

    //--------------------------------------------------------------------
    // Implements ActionListener
    //--------------------------------------------------------------------
    public void actionPerformed(ActionEvent e) {
        try {
            JMenuItem source = (JMenuItem) e.getSource();
            if (source == miFileOpenDataXVG) {
                openDataXvg();
            } else if (source == miFileOpenDataXPM) {
                openDataXpm();
            } else if (source == miFileOpenDataPCA) {
                openDataPca();
            } else if (source == miFileExit) {
                exit();
            } else if (source == miXaxisFrame) {
                ampXaxis(0);
            } else if (source == miXaxis100) {
                ampXaxis(100);
            } else if (source == miXaxis200) {
                ampXaxis(200);
            } else if (source == miXaxis400) {
                ampXaxis(400);
            } else if (source == miXaxis800) {
                ampXaxis(800);
            } else if (source == miXaxis1600) {
                ampXaxis(1600);
            } else if (source == miYaxisFrame) {
                ampYaxis(0);
            } else if (source == miYaxis100) {
                ampYaxis(100);
            } else if (source == miYaxis200) {
                ampYaxis(200);
            } else if (source == miYaxis400) {
                ampYaxis(400);
            } else if (source == miYaxis800) {
                ampYaxis(800);
            } else if (source == miYaxis1600) {
                ampYaxis(1600);
            } else if (source == miYaxisRange) {
                changeYaxisRange();
            }

        } catch (CustomException ex) {
            handleCustomException(ex);
        } catch (Exception ex) {
            handleException(ex);
        }
    }

    //--------------------------------------------------------------------
    // Refresh image
    //--------------------------------------------------------------------
    private void refreshImage() {
        refreshImage(0, 0);
    }

    private void refreshImage(double rw, double rh) {
        if (data != null) {
            int bw = scrollPane.getHorizontalScrollBar().getValue();
            int bh = scrollPane.getVerticalScrollBar().getValue();
            int sw = scrollPane.getWidth();
            int sh = scrollPane.getHeight();
            int w = (plotWidth == 0) ? sw : plotWidth;
            int h = (plotHeight == 0) ? sh : plotHeight;
            context.setWidth(w);
            context.setHeight(h - 15);
            plotPanel.setPreferredSize(new Dimension(w, h));
            BufferedImage image = renderer.render(data);
            plotPanel.setImage(image);
            plotPanel.repaint();
            if (rw > 0 && rh > 0) {
                // enforce refresh
                scrollPane.setViewportView(plotPanel);
                scrollPane.getHorizontalScrollBar().setValue(
                        adjustScrollBar(bw, rw, sw, w));
                scrollPane.getVerticalScrollBar().setValue(
                        adjustScrollBar(bh, rh, sh, h));
            }
        }
    }

    // try to keep center
    private int adjustScrollBar(int b, double r, int d, int p) {
        int v = (int) (r * b + (r - 1) * d / 2);
        if (v < 0) {
            return 0;
        } else if (v > p - d) {
            return p - d;
        } else {
            return v;
        }
    }

    //--------------------------------------------------------------------
    // Files
    //--------------------------------------------------------------------
    public void openDataXvg(File file) throws CustomException {
        String title = file.getName();
        if (title.lastIndexOf(".") != -1) {
            title = title.substring(0, title.lastIndexOf("."));
        }
        setTitle(title);

        DataFileReaderXvg reader = new DataFileReaderXvg();
        data = reader.read(file);
        renderer = new ImageRendererLine(context);

        scaleWithFrame();
        refreshImage();
        miYaxisRange.setEnabled(true);
    }

    public void openDataXpm(File file) throws CustomException {
        String title = file.getName();
        if (title.lastIndexOf(".") != -1) {
            title = title.substring(0, title.lastIndexOf("."));
        }
        setTitle(title);

        DataFileReaderXpm reader = new DataFileReaderXpm();
        data = reader.read(file);
        renderer = new ImageRendererTile(context, false);

        scaleWithFrame();
        refreshImage();
        miYaxisRange.setEnabled(false);
    }
    
    public void openDataPca(File file, float start, float step)
            throws CustomException {
        String title = file.getName();
        if (title.lastIndexOf(".") != -1) {
            title = title.substring(0, title.lastIndexOf("."));
        }
        setTitle(title);

        DataFileReaderPca reader = new DataFileReaderPca();
        data = reader.read(file, start, step);
        renderer = new ImageRendererTile(context, true);

        scaleWithFrame();
        refreshImage();
        miYaxisRange.setEnabled(false);
    }

    private void openDataXvg() throws CustomException {
        File file = chooseFile();
        if (file != null) {
            openDataXvg(file);
        }        
    }

    private void openDataXpm() throws CustomException {
        File file = chooseFile();
        if (file != null) {
            openDataXpm(file);
        }        
    }
    
    private void openDataPca() throws CustomException {
        File file = chooseFile();
        if (file != null) {
            openDataPca(file, 10, 10);
        }        
    }

    private File chooseFile() {
        JFileChooser jFileChooser = new JFileChooser(currDir);
        jFileChooser.setMultiSelectionEnabled(false);
        if (jFileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
            File file = jFileChooser.getSelectedFile();
            currDir = file.getParentFile();
            return file;
        }
        return null;
    }

    //--------------------------------------------------------------------
    // Menu X-axis Y-axis
    //--------------------------------------------------------------------
    public void ampXaxis(int amp) {
        int prev = (plotWidth == 0) ? scrollPane.getWidth() : plotWidth;
        plotWidth = PANEL_WIDTH * amp / 100;
        double rw = plotWidth / (double) prev;
        if (plotWidth == 0) {
            scrollPane.setHorizontalScrollBarPolicy(
                    JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            scrollPane.setHorizontalScrollBarPolicy(
                    JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        }
        refreshImage(rw, 1);
    }
    
    public void ampYaxis(int amp) {
        int prev = (plotHeight == 0) ? scrollPane.getHeight() : plotHeight;
        plotHeight = PANEL_HEIGHT * amp / 100;
        double rh = plotHeight / (double) prev;
        if (plotHeight == 0) {
            scrollPane.setVerticalScrollBarPolicy(
                    JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        } else {
            scrollPane.setVerticalScrollBarPolicy(
                    JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        }
        refreshImage(1, rh);
    }

    public void changeYaxisRange() {
        if (data != null) {
            PlotDataXvg dataXvg = (PlotDataXvg) data;
            if (yaxisRangeDialog == null) {
                yaxisRangeDialog = new YaxisRangeDialog(this, "Y-axis range", true);
            }
            yaxisRangeDialog.initialize();
            yaxisRangeDialog.setMin(dataXvg.getViewMinY());
            yaxisRangeDialog.setMax(dataXvg.getViewMaxY());
            yaxisRangeDialog.setVisible(true);
            if (yaxisRangeDialog.isOk()) {
                float min = yaxisRangeDialog.getMin();
                float max = yaxisRangeDialog.getMax();
                dataXvg.recreateYlabelGrid(min, max);
                refreshImage();
            }
            if (yaxisRangeDialog.isReset()) {
                dataXvg.createLabelGrid();
                refreshImage();
            }
        }
    }

    //--------------------------------------------------------------------
    // Exceptions
    //--------------------------------------------------------------------
    private void handleCustomException(CustomException e) {
        JOptionPane.showMessageDialog(this, e.getMessage(), "Warning",
                JOptionPane.WARNING_MESSAGE);
    }

    private void handleException(Exception e) {
        String message = "Unexpected error has occurred.\n";
        if (e.getMessage() != null) {
            message += "\n" + e.getMessage();
        }
        JOptionPane.showMessageDialog(this, message, "Error",
                JOptionPane.ERROR_MESSAGE);
        e.printStackTrace();
        exit();
    }

    //--------------------------------------------------------------------
    // Creates components
    //--------------------------------------------------------------------
    private void createComponents() {
        plotPanel = new PlotPanel(this);
        scrollPane = new JScrollPane(plotPanel);
        scrollPane.addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                refreshImage();
            }
        });
        scaleWithFrame();

        Container container = getContentPane();
        container.setLayout(new BorderLayout());
        container.add(scrollPane, BorderLayout.CENTER);
    }

    private void scaleWithFrame() {
        plotWidth = 0;
        plotHeight = 0;
        scrollPane.setVerticalScrollBarPolicy(
                JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        scrollPane.setHorizontalScrollBarPolicy(
                JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        miXaxisFrame.setSelected(true);
        miYaxisFrame.setSelected(true);
    }

    //--------------------------------------------------------------------
    // Creates menu
    //--------------------------------------------------------------------
    private void createMenu() {
        // File
        JMenu menuFile = new JMenu("File");
        miFileOpenDataXVG = createMenuItem("Open XVG Data");
        miFileOpenDataXPM = createMenuItem("Open XPM Data");
        miFileOpenDataPCA = createMenuItem("Open PCA Data");
        miFileExit = createMenuItem("Exit");
        menuFile.add(miFileOpenDataXVG);
        menuFile.add(miFileOpenDataXPM);
        menuFile.add(miFileOpenDataPCA);
        menuFile.addSeparator();
        menuFile.add(miFileExit);
        
        // X-axis
        JMenu menuXaxis= new JMenu("X-axis");
        miXaxisFrame = createCheckBoxMenuItem("frame width");
        miXaxis100 = createCheckBoxMenuItem(
                "100% (" + PANEL_WIDTH + " pixel)");
        miXaxis200 = createCheckBoxMenuItem("200%");
        miXaxis400 = createCheckBoxMenuItem("400%");
        miXaxis800 = createCheckBoxMenuItem("800%");
        miXaxis1600 = createCheckBoxMenuItem("1600%");
        ButtonGroup groupXaxis = new ButtonGroup();
        groupXaxis.add(miXaxisFrame);
        groupXaxis.add(miXaxis100);
        groupXaxis.add(miXaxis200);
        groupXaxis.add(miXaxis400);
        groupXaxis.add(miXaxis800);
        groupXaxis.add(miXaxis1600);
        menuXaxis.add(miXaxisFrame);
        menuXaxis.add(miXaxis100);
        menuXaxis.add(miXaxis200);
        menuXaxis.add(miXaxis400);
        menuXaxis.add(miXaxis800);
        menuXaxis.add(miXaxis1600);
        
        // Y-axis
        JMenu menuYaxis= new JMenu("Y-axis");
        miYaxisFrame = createCheckBoxMenuItem("frame height");
        miYaxis100 = createCheckBoxMenuItem(
                "100% (" + PANEL_HEIGHT + " pixel)");
        miYaxis200 = createCheckBoxMenuItem("200%");
        miYaxis400 = createCheckBoxMenuItem("400%");
        miYaxis800 = createCheckBoxMenuItem("800%");
        miYaxis1600 = createCheckBoxMenuItem("1600%");
        miYaxisRange = createMenuItem("range");
        ButtonGroup groupYaxis = new ButtonGroup();
        groupYaxis.add(miYaxisFrame);
        groupYaxis.add(miYaxis100);
        groupYaxis.add(miYaxis200);
        groupYaxis.add(miYaxis400);
        groupYaxis.add(miYaxis800);
        groupYaxis.add(miYaxis1600);
        menuYaxis.add(miYaxisFrame);
        menuYaxis.add(miYaxis100);
        menuYaxis.add(miYaxis200);
        menuYaxis.add(miYaxis400);
        menuYaxis.add(miYaxis800);
        menuYaxis.add(miYaxis1600);
        menuYaxis.addSeparator();
        menuYaxis.add(miYaxisRange);
        
        // menu bar
        JMenuBar menuBar = new JMenuBar();
        if (standalone) {
            menuBar.add(menuFile);
            menuBar.add(Box.createRigidArea(new Dimension(20, 1)));
        }
        menuBar.add(menuXaxis);
        menuBar.add(Box.createRigidArea(new Dimension(20, 1)));
        menuBar.add(menuYaxis);
        setJMenuBar(menuBar);
    }

    private JMenuItem createMenuItem(String label) {
        JMenuItem menuItem = new JMenuItem(label);
        menuItem.addActionListener(this);
        return menuItem;
    }
    
    private JRadioButtonMenuItem createCheckBoxMenuItem(String label) {
        JRadioButtonMenuItem menuItem = new JRadioButtonMenuItem(label);
        menuItem.addActionListener(this);
        return menuItem;
    }

    public static void main(String[] args) {
        Frame frame = new Plot2dFrame(true);
        frame.validate();

        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        int maxHeight = (int)(0.9f*screenSize.height);
        int maxWidth = (int)(0.9f*screenSize.width);
        int maxSize = (maxHeight < maxWidth) ? maxHeight : maxWidth;
        
        Dimension frameSize = frame.getSize();
        if(frameSize.height>maxSize || frameSize.width>maxSize){
            frameSize.height = maxSize;
            frameSize.width = maxSize;
            frame.setSize(frameSize);
        }
        
        frame.setLocation(
            (screenSize.width - frameSize.width)/2,
            (screenSize.height - frameSize.height)/2
        );

        frame.setVisible(true);
    }
}
