package jp.riken.plot2d.renderer;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.text.*;
import jp.riken.plot2d.data.*;

public abstract class ImageRenderer {

    protected RenderingContext context;

    // internal use
    protected int w;
    protected int h;
    protected double x1;
    protected double x2;
    protected double y1;
    protected double y2;
    protected double minX;
    protected double maxX;
    protected double minY;
    protected double maxY;
    protected double sx;
    protected double sy;
    protected double ratio;

    protected NumberFormat formatter;

    public ImageRenderer(RenderingContext context) {
        this.context = context;
        formatter = NumberFormat.getNumberInstance();
        formatter.setMaximumFractionDigits(4);
        formatter.setMinimumFractionDigits(4);
        formatter.setGroupingUsed(false);
    }

    public BufferedImage render(PlotData data) {
        w = context.getWidth();
        h = context.getHeight();
        BufferedImage image = new BufferedImage(
                w, h, BufferedImage.TYPE_INT_RGB);
        render(data, image.createGraphics());
        return image;
    }

    protected abstract void render(PlotData data, Graphics2D g);

    // --------------------------------------------------------------------
    // scale
    // --------------------------------------------------------------------
    protected void readScale(Grid grid) {
        x1 = context.getImageInitX();
        x2 = context.getImageTermX();
        y1 = context.getImageInitY();
        y2 = context.getImageTermY();
        minX = grid.getMinX();
        maxX = grid.getMaxX();
        minY = grid.getMinY();
        maxY = grid.getMaxY();
        sx = (x2 - x1) / (maxX - minX);
        sy = (y2 - y1) / (maxY - minY);
        double rw = w / 700.0;
        double rh = h / 560.0;
        ratio = (rw < rh) ? rw : rh;
    }
    
    protected void readScale(Grid grid, int pca) {
        x1 = context.getImageInitX();
        x2 = context.getImageTermX();
        y1 = context.getImageInitY();
        y2 = context.getImageTermY();
        
        double bx = x2 * 0.46;
        double by = y2 * 0.46;
        double bx1 = bx - x1;
        double by1 = by - y1;
        
        if (pca == 0) {
            x2 = bx;
            y1 = y2 - by1;
        } else if (pca == 1) {
            x1 = x2 - bx1;
            y1 = y2 - by1;
        } else if (pca == 2) {
            x2 = bx;
            y2 = by;
        } else if (pca == 3) {
            x1 = x2 - bx1;
            y2 = by;
        }
        
        minX = grid.getMinX();
        maxX = grid.getMaxX();
        minY = grid.getMinY();
        maxY = grid.getMaxY();
        sx = (x2 - x1) / (maxX - minX);
        sy = (y2 - y1) / (maxY - minY);
        double rw = w / 700.0;
        double rh = h / 560.0;
        ratio = (rw < rh) ? rw : rh;
    }

    // --------------------------------------------------------------------
    // background
    // --------------------------------------------------------------------
    protected void renderBackground(Graphics2D g) {
        g.setColor(context.getBackgroundColor());
        g.fillRect(0, 0, w, h);
    }

    // --------------------------------------------------------------------
    // frame
    // --------------------------------------------------------------------
    protected void renderFrame(Graphics2D g) {
        g.setColor(context.getFrameColor());
        g.drawRect((int) x1, (int) (h - y2),
                (int) (x2 - x1) + 1, (int) (y2 - y1) + 1);
    }

    // --------------------------------------------------------------------
    // title
    // --------------------------------------------------------------------
    protected void renderTitle(PlotData data, Graphics2D g) {
        int fontSize2 = (int) (24 * ratio);
        Font font2 = new Font("Dialog", Font.PLAIN, fontSize2);
        FontMetrics fm2 = g.getFontMetrics(font2);

        String title = data.getTitle();

        int xp = (int) ((x2 - x1) / 2 + x1 - 0.5 * fm2.stringWidth(title));
        int yp = (int) (h - y2 - 0.9 * fm2.getAscent());

        g.setColor(Color.BLACK);
        g.setFont(font2);
        g.drawString(title, xp, yp);

        // draw subtitle
        int fontSize1 = (int) (14 * ratio);
        Font font1 = new Font("Dialog", Font.PLAIN, fontSize1);
        FontMetrics fm1 = g.getFontMetrics(font1);

        String subtitle = data.getSubtitle();

        xp = (int) ((x2 - x1) / 2 + x1 - 0.5 * fm1.stringWidth(subtitle));
        yp = (int) (h - y2 - 0.2 * fm2.getAscent());

        g.setColor(Color.BLACK);
        g.setFont(font1);
        g.drawString(subtitle, xp, yp);
    }

    // --------------------------------------------------------------------
    // label
    // --------------------------------------------------------------------
    protected void renderLabel(PlotData data, Graphics2D g, boolean pca) {
        g.setColor(Color.BLACK);

        Grid labelGrid = data.getLabelGrid();
        double dx = labelGrid.getWidthX();
        double dy = labelGrid.getWidthY();
        int numX = labelGrid.getCountX();
        int numY = labelGrid.getCountY();
        
        int fontSize1 = 0;
        int fontSize2 = 0;
        if (pca) {
            fontSize1 = (int) (10 * ratio);
            fontSize2 = (int) (18 * ratio);
        }
        else {
            fontSize1 = (int) (14 * ratio);
            fontSize2 = (int) (24 * ratio);
        }

        Font font1 = new Font("Dialog", Font.PLAIN, fontSize1);
        Font font2 = new Font("Dialog", Font.PLAIN, fontSize2);

        FontMetrics fm1 = g.getFontMetrics(font1);
        FontMetrics fm2 = g.getFontMetrics(font2);

        int ascent1 = fm1.getAscent();
        int ascent2 = fm2.getAscent();

        int xp = 0;
        int yp = 0;

        // graduations x
        g.setFont(font1);
        yp = (int) (h - y1 + 1.5 * ascent1);
        for (int i = 0; i <= numX; i++) {
            double x = i * dx;
            String value = formatValue(minX + x);
            xp = (int) (x * sx + x1 - fm1.stringWidth(value) / 2);
            g.drawString(value, xp, yp);
        }

        // label x
        String labelX = data.getLabelX();
        xp = (int) (numX * dx * sx / 2 + x1 - fm2.stringWidth(labelX) / 2);
        yp += ascent2;
        g.setFont(font2);
        g.drawString(labelX, xp, yp);

        // graduations y
        g.setFont(font1);
        int xp2 = Integer.MAX_VALUE;
        for (int i = 0; i <= numY; i++) {
            double y = i * dy;
            String value = formatValue(minY + y);
            xp = (int) (x1 - fm1.stringWidth(value) - 0.5 * ascent1);
            yp = (int) (h - y1 - y * sy + 0.5 * ascent1);
            if (xp < xp2) {
                xp2 = xp;
            }
            g.drawString(value, xp, yp);
        }

        // label y
        String labelY = data.getLabelY();

        xp2 -= 0.7 * ascent2;
        int yp2 = (int) (h - numY * dy * sy / 2 - y1 +
                fm2.stringWidth(labelY) / 2);

        AffineTransform transform = new AffineTransform();
        transform.translate(xp2, yp2);
        transform.rotate(-Math.PI / 2);
        g.setTransform(transform);

        g.setFont(font2);
        g.drawString(labelY, 0, 0);

        transform.setToIdentity();
        g.setTransform(transform);
    }

    // --------------------------------------------------------------------
    // picked line
    // --------------------------------------------------------------------
    protected void renderPickedLine(PlotData data, Graphics2D g) {
        if (data.getPickValueX() != null) {
            double x = data.getPickValueX();
            if (x >= minX && x <= maxX) {
                int xp = (int) ((x - minX) * sx + x1);
                int yp1 = (int) (h - y1);
                int yp2 = (int) (h - y2) + 1;
                g.setColor(new Color(225, 0, 178));
                g.drawLine(xp, yp1, xp, yp2);
            }
        }
    }

    // --------------------------------------------------------------------
    // utility
    // --------------------------------------------------------------------
    protected String formatValue(double d) {
        String value = formatter.format(d);
        if (value.indexOf(".") == -1) {
            return value;
        }
        int index = value.length() - 1;
        char curr = value.charAt(index);
        while (curr == '0') {
            curr = value.charAt(--index);
        }
        if (curr == '.') {
            index--;
        }
        return value.substring(0, index + 1);
    }
}
