package jp.riken.viewer.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.ArrayList;
import java.util.List;
import javax.swing.*;
import jp.riken.common.api.*;
import jp.riken.common.exception.*;
import jp.riken.plot2d.data.*;
import jp.riken.plot2d.gui.*;
import jp.riken.viewer.gromacs.*;

public class ApplicationFrame extends ViewerFrame implements Viewer {

    private static final long serialVersionUID = 1L;

    // data
    private float commonTime;

    // component
    private TrajectoryControlPanel trajectoryControlPanel;
    private OpenTrajectoryDialog openTrajectoryDialog;
    private OpenPcaDialog openPcaDialog;
    private OpenHbondDialog openHbondDialog;
    private CommandFrame commandFrame;
    private List<Plot2dFrame> plot2dFrames;

    // menu
    private MenuItem miFileOpenTrajectory;
    private MenuItem miFileOpenXvg;
    private MenuItem miFileOpenXpm;
    private MenuItem miFileOpenPca;
    private MenuItem miFileOpenDssp;
    private MenuItem miFileOpenHbond;
    private MenuItem miFileOpenScript;
    private MenuItem miFileCloseAll;
    private MenuItem miFileCloseTrajectory;
    private MenuItem miFileSavePng;
    private MenuItem miFileSaveJpeg;
    private MenuItem miFileExit;
    private MenuItem miDisplayProteinWireframe;
    private MenuItem miDisplayProteinBackbone;
    private MenuItem miDisplayProteinStick;
    private MenuItem miDisplayProteinSpacefill;
    private MenuItem miDisplayProteinBallStick;
    private MenuItem miDisplayProteinRibbon;
    private MenuItem miDisplayProteinCartoon;
    private MenuItem miDisplayLigandWireframe;
    private MenuItem miDisplayLigandStick;
    private MenuItem miDisplayLigandSpacefill;
    private MenuItem miDisplayLigandBallStick;
    private MenuItem miDisplayLigandNeighborWireframe;
    private MenuItem miDisplayLigandNeighborStick;
    private MenuItem miDisplayLigandNeighborSpacefill;
    private MenuItem miDisplayLigandNeighborBallStick;
    private MenuItem miDisplayLigandNeighborOff;
    private MenuItem miDisplayLigandHbondDashedLine;
    private MenuItem miDisplayLigandHbondStick;
    private MenuItem miDisplayLigandHbondOff;
    private MenuItem miDisplaySolventWireframe;
    private MenuItem miDisplaySolventOff;
    private MenuItem miOptionCommandLine;

    // file chooser directory
    private File currDir;

    public ApplicationFrame() {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            SwingUtilities.updateComponentTreeUI(this);
        } catch (Exception e) {
        }

        if (controller == null) {
            return;
        }

        // WindowListner
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                exit();
            }

            public void windowOpened(WindowEvent e) {
                init();
            }
        });

        // size & title
        setSize(new Dimension(700, 700));
        setTitle("");

        createComponent();
        createMenu();
        updateMenu();
        plot2dFrames = new ArrayList<Plot2dFrame>();

        currDir = new File(".");
    }

    private void init() {
        commandArea.setActive(true);
        commandArea.requestFocus();
        commandQueue.setActive(true);
        commandThread.start();
    }

    public void dispose() {
        if (controller != null) {
            commandThread.terminate();
            try {
                commandThread.join();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }

            if (commandFrame != null) {
                commandFrame.dispose();
            }
            for (Plot2dFrame plot2d : plot2dFrames) {
                plot2d.dispose();
            }
            controller.dispose();
        }
        if (openTrajectoryDialog != null) {
            openTrajectoryDialog.dispose();
        }
        if (openPcaDialog != null) {
            openPcaDialog.dispose();
        }
        if (openHbondDialog != null) {
            openHbondDialog.dispose();
        }
        super.dispose();
    }

    // --------------------------------------------------------------------
    // Implements Viewer
    // --------------------------------------------------------------------
    @Override
    public void setTime(float time) {
        if (moleculeFile != null) {
            int frame = trajectoryControlPanel.findNearestFrame(time);
            setTrajectoryFrame(frame);
        }
        commonTime = time;
        sendTime();
    }

    @Override
    public void removePlot2d(Plot2d plot2d) {
        plot2dFrames.remove(plot2d);
    }

    // --------------------------------------------------------------------
    // Trajectory
    // --------------------------------------------------------------------
    public void updateTrajectoryFrame(int frame) {
        if (moleculeFile != null) {
            setTrajectoryFrame(frame);
            commonTime = moleculeFile.getMoleculeTraj(frame).getTime();
            sendTime();
        }
    }

    private void setTrajectoryFrame(int frame) {
        moleculeFile.setCurrentFrame(frame);
        controller.refresh();
        trajectoryControlPanel.setFrame(frame);
    }

    private void sendTime() {
        for (Plot2d plot2d : plot2dFrames) {
            plot2d.setTime(commonTime);
        }
    }

    // --------------------------------------------------------------------
    // File
    // --------------------------------------------------------------------
    public File chooseFile(GromacsFileType fileType, boolean save) {
        JFileChooser jFileChooser = new JFileChooser(currDir);
        jFileChooser.setMultiSelectionEnabled(false);
        if (fileType != null) {
            jFileChooser.setFileFilter(fileType);
        }

        int option;
        if (save) {
            option = jFileChooser.showSaveDialog(this);
        } else {
            option = jFileChooser.showOpenDialog(this);
        }

        if (option == JFileChooser.APPROVE_OPTION) {
            File file = jFileChooser.getSelectedFile();
            currDir = file.getParentFile();
            return file;
        }
        return null;
    }

    private void loadTrajectory() {
        try {
            File reference = openTrajectoryDialog.getReferenceFile();
            File trajectory = openTrajectoryDialog.getTrajectoryFile();
            TrajectoryReaderOption option = openTrajectoryDialog.getOption();
            GromacsMoleculeReader reader = null;
            String[] items = reference.getName().split("\\.");
            if (items[items.length - 1].equalsIgnoreCase("gro")) {
                reader = new GromacsGroReader();
            } else {
                reader = new GromacsPdbReader();
            }
            moleculeFile = reader.readAsReference(reference, option);
            moleculeFile.setAnimation(true);
            GromacsTrajectoryReader trajReader = new GromacsTrajectoryReader();
            trajReader.read(trajectory, moleculeFile, option);

            controller.loadMoleculeFile(moleculeFile, true, null);
            trajectoryControlPanel.fileOpened(moleculeFile);
            updateTrajectoryFrame(0);

        } catch (Exception e) {
            throw new CustomException(e);
        }
    }

    private void closeTrajectory() {
        controller.closeAllFile();
        trajectoryControlPanel.fileClosed();
        moleculeFile = null;
    }

    // --------------------------------------------------------------------
    // Component
    // --------------------------------------------------------------------
    private void createComponent() {
        trajectoryControlPanel = new TrajectoryControlPanel(this);
        Container container = getContentPane();
        container.setLayout(new BorderLayout());
        container.add(controller.getCanvasL(), BorderLayout.CENTER);
        container.add(trajectoryControlPanel, BorderLayout.SOUTH);
    }

    private void showOpenTrajectoryDialog() {
        if (openTrajectoryDialog == null) {
            openTrajectoryDialog = new OpenTrajectoryDialog(this,
                    "Open Trajectory", true);
        }
        openTrajectoryDialog.initialize();
        openTrajectoryDialog.setVisible(true);
    }

    private void showOpenPcaDialog() {
        if (openPcaDialog == null) {
            openPcaDialog = new OpenPcaDialog(this, "Open PCA", true);
        }
        openPcaDialog.initialize();
        openPcaDialog.setVisible(true);
    }

    private void showOpenHbondDialog() {
        if (openHbondDialog == null) {
            openHbondDialog = new OpenHbondDialog(this, "Open H-bond", true);
        }
        openHbondDialog.initialize();
        openHbondDialog.setVisible(true);
    }

    private Plot2dFrame createAndShowPlot2dFrame() {
        Plot2dFrame plot2d = new Plot2dFrame(false);
        plot2dFrames.add(plot2d);
        plot2d.setViewer(this);
        plot2d.setCenterTo(this);
        plot2d.setVisible(true);
        return plot2d;
    }

    // --------------------------------------------------------------------
    // Menu
    // --------------------------------------------------------------------
    private void createMenu() {
        Menu menuFile = new Menu("File");
        Menu menuFileOpen = new Menu("Open");
        miFileOpenTrajectory = createMenuItem("Trajectory");
        miFileOpenXvg = createMenuItem("XVG");
        miFileOpenXpm = createMenuItem("XPM");
        miFileOpenPca = createMenuItem("PCA");
        miFileOpenDssp = createMenuItem("DSSP");
        miFileOpenHbond = createMenuItem("H-bond");
        miFileOpenScript = createMenuItem("Script");
        menuFileOpen.add(miFileOpenTrajectory);
        menuFileOpen.add(miFileOpenXvg);
        menuFileOpen.add(miFileOpenXpm);
        menuFileOpen.add(miFileOpenPca);
        menuFileOpen.addSeparator();
        menuFileOpen.add(miFileOpenDssp);
        menuFileOpen.add(miFileOpenHbond);
        menuFileOpen.addSeparator();
        menuFileOpen.add(miFileOpenScript);
        menuFile.add(menuFileOpen);
        Menu menuFileClose = new Menu("Close");
        miFileCloseAll = createMenuItem("All");
        miFileCloseTrajectory = createMenuItem("Trajectory");
        menuFileClose.add(miFileCloseAll);
        menuFileClose.add(miFileCloseTrajectory);
        menuFile.add(menuFileClose);
        Menu menuFileSave = new Menu("Save");
        miFileSavePng = createMenuItem("PNG");
        miFileSaveJpeg = createMenuItem("JPEG");
        menuFileSave.add(miFileSavePng);
        menuFileSave.add(miFileSaveJpeg);
        menuFile.add(menuFileSave);
        menuFile.addSeparator();
        miFileExit = createMenuItem("Exit");
        menuFile.add(miFileExit);

        Menu menuDisplay = new Menu("Display");
        Menu menuProtein = new Menu("Protein");
        miDisplayProteinWireframe = createMenuItem("Wireframe");
        miDisplayProteinBackbone = createMenuItem("Backbone");
        miDisplayProteinStick = createMenuItem("Stick");
        miDisplayProteinSpacefill = createMenuItem("Spacefill");
        miDisplayProteinBallStick = createMenuItem("Ball & Stick");
        miDisplayProteinRibbon = createMenuItem("Ribbon");
        miDisplayProteinCartoon = createMenuItem("Cartoon");
        menuProtein.add(miDisplayProteinWireframe);
        menuProtein.add(miDisplayProteinBackbone);
        menuProtein.add(miDisplayProteinStick);
        menuProtein.add(miDisplayProteinSpacefill);
        menuProtein.add(miDisplayProteinBallStick);
        menuProtein.add(miDisplayProteinRibbon);
        menuProtein.add(miDisplayProteinCartoon);
        menuDisplay.add(menuProtein);
        Menu menuLigand = new Menu("Ligand");
        miDisplayLigandWireframe = createMenuItem("Wireframe");
        miDisplayLigandStick = createMenuItem("Stick");
        miDisplayLigandSpacefill = createMenuItem("Spacefill");
        miDisplayLigandBallStick = createMenuItem("Ball & Stick");
        menuLigand.add(miDisplayLigandWireframe);
        menuLigand.add(miDisplayLigandStick);
        menuLigand.add(miDisplayLigandSpacefill);
        menuLigand.add(miDisplayLigandBallStick);
        menuDisplay.add(menuLigand);
        Menu menuLigandNeighbor = new Menu("Ligand Neighbor");
        miDisplayLigandNeighborWireframe = createMenuItem("Wireframe");
        miDisplayLigandNeighborStick = createMenuItem("Stick");
        miDisplayLigandNeighborSpacefill = createMenuItem("Spacefill");
        miDisplayLigandNeighborBallStick = createMenuItem("Ball & Stick");
        miDisplayLigandNeighborOff = createMenuItem("Off");
        menuLigandNeighbor.add(miDisplayLigandNeighborWireframe);
        menuLigandNeighbor.add(miDisplayLigandNeighborStick);
        menuLigandNeighbor.add(miDisplayLigandNeighborSpacefill);
        menuLigandNeighbor.add(miDisplayLigandNeighborBallStick);
        menuLigandNeighbor.add(miDisplayLigandNeighborOff);
        menuDisplay.add(menuLigandNeighbor);
        Menu menuLigandHbond = new Menu("Ligand H-bond");
        miDisplayLigandHbondDashedLine = createMenuItem("Dashed Line");
        miDisplayLigandHbondStick = createMenuItem("Stick");
        miDisplayLigandHbondOff = createMenuItem("Off");
        menuLigandHbond.add(miDisplayLigandHbondDashedLine);
        menuLigandHbond.add(miDisplayLigandHbondStick);
        menuLigandHbond.add(miDisplayLigandHbondOff);
        menuDisplay.add(menuLigandHbond);
        Menu menuSolvent = new Menu("Solvent");
        miDisplaySolventWireframe = createMenuItem("Wireframe");
        miDisplaySolventOff = createMenuItem("Off");
        menuSolvent.add(miDisplaySolventWireframe);
        menuSolvent.add(miDisplaySolventOff);
        menuDisplay.add(menuSolvent);

        Menu menuOption = new Menu("Option");
        miOptionCommandLine = createMenuItem("Command Line");
        menuOption.add(miOptionCommandLine);

        MenuBar menuBar = new MenuBar();
        menuBar.add(menuFile);
        menuBar.add(menuDisplay);
        menuBar.add(menuOption);
        setMenuBar(menuBar);
    }

    private MenuItem createMenuItem(String label) {
        MenuItem menuItem = new MenuItem(label);
        menuItem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                menuActionPerformed(e);
            }
        });
        return menuItem;
    }

    private void menuActionPerformed(ActionEvent e) {
        MenuItem item = (MenuItem)e.getSource();
        try {
            if (item == miFileOpenTrajectory) {
                showOpenTrajectoryDialog();
                if (openTrajectoryDialog.isOk()) {
                    loadTrajectory();
                }
            } else if (item == miFileOpenXvg) {
                File file = chooseFile(GromacsFileType.XVG, false);
                if (file != null) {
                    Plot2dFrame plot2d = createAndShowPlot2dFrame();
                    plot2d.openDataXvg(file);
                    sendTime();
                }
            } else if (item == miFileOpenXpm) {
                File file = chooseFile(GromacsFileType.XPM, false);
                if (file != null) {
                    Plot2dFrame plot2d = createAndShowPlot2dFrame();
                    plot2d.openDataXpm(file);
                    sendTime();
                }
            } else if (item == miFileOpenPca) {
                showOpenPcaDialog();
                if (openPcaDialog.isOk()) {
                    File file = openPcaDialog.getPcaFile();
                    float start = openPcaDialog.getStart();
                    float step = openPcaDialog.getStep();
                    Plot2dFrame plot2d = createAndShowPlot2dFrame();
                    plot2d.openDataPca(file, start, step);
                    sendTime();
                }
            } else if (item == miFileOpenDssp) {
                File file = chooseFile(GromacsFileType.XPM, false);
                if (file != null) {
                    Plot2dFrame plot2d = createAndShowPlot2dFrame();
                    plot2d.openDataXpm(file);
                    processDssp((PlotDataXpm) plot2d.getData());
                    sendTime();
                }
            } else if (item == miFileOpenHbond) {
                showOpenHbondDialog();
                if (openHbondDialog.isOk()) {
                    File indexFile = openHbondDialog.getIndexFile();
                    File mapFile = openHbondDialog.getMapFile();
                    Plot2dFrame plot2d = createAndShowPlot2dFrame();
                    plot2d.openDataXpm(mapFile);
                    processHbond(indexFile, (PlotDataXpm) plot2d.getData());
                    sendTime();
                }
            } else if (item == miFileOpenScript) {
                File file = chooseFile(null, false);
                if (file != null) {
                    loadScript(file);
                }
            } else if (item == miFileCloseAll) {
                for (Plot2dFrame plot2d : plot2dFrames) {
                    plot2d.dispose();
                }
                plot2dFrames.clear();
                closeTrajectory();
            } else if (item == miFileCloseTrajectory) {
                closeTrajectory();
            } else if (item == miFileSavePng) {
                File file = chooseFile(GromacsFileType.PNG, true);
                if (file != null) {
                    savePng(file);
                }
            } else if (item == miFileSaveJpeg) {
                File file = chooseFile(GromacsFileType.JPG, true);
                if (file != null) {
                    saveJpeg(file);
                }
            } else if (item == miFileExit) {
                exit();
            } else if (item == miDisplayProteinWireframe) {
                displayProteinWireframe();
            } else if (item == miDisplayProteinBackbone) {
                displayProteinBackbone();
            } else if (item == miDisplayProteinStick) {
                displayProteinStick();
            } else if (item == miDisplayProteinSpacefill) {
                displayProteinSpacefill();
            } else if (item == miDisplayProteinBallStick) {
                displayProteinBallStick();
            } else if (item == miDisplayProteinRibbon) {
                displayProteinRibbon();
            } else if (item == miDisplayProteinCartoon) {
                displayProteinCartoon();
            } else if (item == miDisplayLigandWireframe) {
                displayLigandWireframe();
            } else if (item == miDisplayLigandStick) {
                displayLigandStick();
            } else if (item == miDisplayLigandSpacefill) {
                displayLigandSpacefill();
            } else if (item == miDisplayLigandBallStick) {
                displayLigandBallStick();
            } else if (item == miDisplayLigandNeighborWireframe) {
                displayLigandNeighborWireframe();
            } else if (item == miDisplayLigandNeighborStick) {
                displayLigandNeighborStick();
            } else if (item == miDisplayLigandNeighborSpacefill) {
                displayLigandNeighborSpacefill();
            } else if (item == miDisplayLigandNeighborBallStick) {
                displayLigandNeighborBallStick();
            } else if (item == miDisplayLigandNeighborOff) {
                displayLigandNeighborOff();
            } else if (item == miDisplayLigandHbondDashedLine) {
                displayLigandHbondDashedLine();
            } else if (item == miDisplayLigandHbondStick) {
                displayLigandHbondStick();
            } else if (item == miDisplayLigandHbondOff) {
                displayLigandHbondOff();
            } else if (item == miDisplaySolventWireframe) {
                displaySolventWireframe();
            } else if (item == miDisplaySolventOff) {
                displaySolventOff();
            } else if (item == miOptionCommandLine) {
                if (commandFrame == null) {
                    commandFrame = new CommandFrame(commandArea);
                    commandFrame.setCenterTo(this);
                }
                commandFrame.setVisible(true);
            }
        } catch (CustomException ex) {
            showWarning(ex);
        } catch (Exception ex) {
            showErrorAndExit(ex);
        }
    }

    @Override
    protected void updateMenu() {
    }

    // --------------------------------------------------------------------
    // Main
    // --------------------------------------------------------------------
    public static void main(String[] args) {
        ApplicationFrame frame = new ApplicationFrame();
        frame.validate();

        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        int maxHeight = (int)(0.9f*screenSize.height);
        int maxWidth = (int)(0.9f*screenSize.width);
        int maxSize = (maxHeight < maxWidth) ? maxHeight : maxWidth;
        
        Dimension frameSize = frame.getSize();
        if(frameSize.height>maxSize || frameSize.width>maxSize){
            frameSize.height = maxSize;
            frameSize.width = maxSize;
            frame.setSize(frameSize);
        }
        
        frame.setLocation(
            (screenSize.width - frameSize.width)/2,
            (screenSize.height - frameSize.height)/2 );

        // start
        if (frame.graphicsReady()) {
            frame.setVisible(true);
        } else {
            JOptionPane.showMessageDialog(frame, "Cannot find JOGL", "Error",
                    JOptionPane.ERROR_MESSAGE);
            frame.dispose();
            System.exit(0);
        }
    }
}
