package jp.riken.viewer.gui;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.media.opengl.GLProfile;
import javax.swing.*;
import jp.riken.common.exception.CustomException;
import jp.riken.plot2d.data.*;
import org.pdbj.viewer.exceptions.*;
import org.pdbj.viewer.utils.*;
import org.pdbj.viewer.writers.ImageFileWriter;
import org.pdbj.viewer.renderer.*;
import org.pdbj.viewer.controller.*;
import org.pdbj.viewer.manipulators.*;
import org.pdbj.viewer.molecules.*;
import org.pdbj.viewer.mouse.*;
import org.pdbj.viewer.commands.*;
import org.pdbj.viewer.gui.*;

public abstract class ViewerFrame extends JFrame implements CommandHandler,
        GuiWrapper, RendererListener, MousePickListener {

    private static final long serialVersionUID = 1L;

    static {
        GLProfile.initSingleton();
    }

    protected MoleculeFileTraj moleculeFile;
    protected ViewerController controller;
    protected MouseBehavior mouseBehavior;
    protected CommandArea commandArea;
    protected CommandQueue commandQueue;
    protected CommandThread commandThread;
    protected CommandExecutor executor;

    private SelectExprParser parser;
    private Context context;

    private boolean screenInitialized;

    public ViewerFrame() {
        try {
            controller = new ViewerController(this, false);
        } catch (Throwable t) {
            t.printStackTrace();
            return;
        }
        controller.stopRenderer();

        commandArea = new CommandArea(this);
        commandArea.setCommandHandler(this);
        mouseBehavior = new MouseBehavior(controller, this);
        mouseBehavior.setMessageArea(commandArea);
        executor = new CommandExecutor(controller, mouseBehavior, this, null);
        executor.setMessageArea(commandArea);
        commandQueue = new CommandQueue();
        commandThread = new CommandThread(this, commandQueue, executor);

        parser = new SelectExprParser();
        context = new Context(controller, mouseBehavior, this, null);
    }

    public boolean graphicsReady() {
        return controller != null && controller.graphicsReady();
    }

    // --------------------------------------------------------------------
    // Implements RendererListener
    // --------------------------------------------------------------------
    public void screenInitialized() {
    }

    public void screenDrawed() {
        if (!screenInitialized) {
            commandQueue.putCommand("",
                    new CommandType(CommandType.START_RENDERER));
            screenInitialized = true;
        }
    }

    public void screenResized(int x, int y, int width, int height) {
    }

    // --------------------------------------------------------------------
    // Implements CommandHandler
    // --------------------------------------------------------------------
    public void executeSingleCommand(String command) {
        commandQueue.putCommand(command);
    }

    public CommandComplement complementCommand(String command) {
        return executor.complement(command);
    }

    // --------------------------------------------------------------------
    // Implements GuiWrapper
    // --------------------------------------------------------------------
    public boolean isScriptMode() {
        return false;
    }

    public void loadPDBFile(String filePath, FileStatus copy) {
    }

    public void loadPDBmlFile(String filePath, FileStatus copy) {
    }

    public void loadPolygonFile(String filePath, FileStatus copy) {
    }

    public void loadPolygonCgoFile(String filePath, FileStatus copy) {
    }

    public void loadScriptFile(String filePath) {
        loadScript(new File(filePath));
    }

    public void loadAnimationFile(String filePath, FileStatus copy) {
    }

    public void loadPDBmlFileFtp(String pdbCode, FileStatus copy) {
    }

    public void loadEfsite(String efsiteId) {
    }

    public void savePDBFile(String filePath) {
    }

    public void saveScriptFile(String filePath) {
    }

    public void savePngFile(String filePath, float ratio) {
        savePng(new File(filePath));
    }

    public void saveJpegFile(String filePath, float ratio, float quality) {
        saveJpeg(new File(filePath));
    }

    public void pauseScript() {
    }

    public void setupStereo() {
    }

    public void exit() {
        dispose();
        System.exit(0);
    }

    public MessageListener getFtpMessageListener() {
        return null;
    }

    public boolean isStandalone() {
        return true;
    }

    public void loadFunctionFtp(FileStatus file) {
    }

    @SuppressWarnings("rawtypes")
    public HashMap getSiteUrlMap() {
        return controller.getSiteUrlMap().getMap();
    }

    public void loadSiteList(FileStatus file, String key) {
    }

    public Dimension getImageSize() {
        return controller.getCanvasL().getSize();
    }

    public Dimension setImageSize(Dimension d) {
        int w = d.width;
        int h = d.height;
        if (w < 10) {
            w = 10;
        }
        if (h < 10) {
            h = 10;
        }
        Dimension imageSize = controller.getCanvasL().getSize();
        int iw = imageSize.width;
        int ih = imageSize.height;
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        int sw = screenSize.width;
        int sh = screenSize.height;
        Dimension frameSize = getSize();
        int fw = frameSize.width;
        int fh = frameSize.height;
        int dw = w - imageSize.width;
        int dh = h - imageSize.height;
        if (fw + dw > sw) {
            dw = sw - fw;
        }
        if (fh + dh > sh) {
            dh = sh - fh;
        }
        setSize(fw + dw, fh + dh);
        setVisible(true);
        controller.refresh();
        return new Dimension(iw + dw, ih + dh);
    }

    public void pdbjExecute(int fileId, String name) {
    }

    public void initializeStereo() {
    }

    public boolean isNewt() {
        return false;
    }

    public ViewerController getViewerController() {
        return controller;
    }

    public MouseBehavior getMouseBehavior() {
        return mouseBehavior;
    }

    public void loadLocalFile(FileType fileType) {
    }

    public void loadRemoteFile(FileType fileType) {
    }

    public void loadRemoteFile(UrlDialog urlDialog) {
    }

    public void loadByPDBID() {
    }

    public void loadByEFsiteID() {
    }

    public void loadHistory(FileType fileType, FileLocator locator) {
    }

    public void addFileHistory(FileType type, FileLocator locator) {
    }

    public void saveFile(FileType fileType) {
    }

    public void showInformation() {
    }

    public void showAnimationFrame() {
    }

    public void showAboutDialog() {
    }

    public void openManual() {
    }

    public void rotateStereo() {
    }

    public void handleRecoverableException(RecoverableException e) {
        showWarning(e);
    }

    public void handleFatalException(FatalException e) {
        showErrorAndExit(e);
    }

    public void startCommand(String command, CommandType type) {
        startCommand(executor.createTokenizer(command), type);
    }

    public void startCommand(CommandTokenizer tokenizer, CommandType type) {
        try {
            if (tokenizer != null) {
                executor.execute(tokenizer);
            }
        } catch (RecoverableException e) {
            handleRecoverableException(e);
        } catch (OutOfMemoryError e) {
            handleFatalException(ExceptionUtil.handleOutOfMemoryError(e));
        } catch (Throwable t) {
            handleFatalException(ExceptionUtil.handleThrowable(t));
        }

        if (type.action == CommandType.START_RENDERER) {
            controller.startRenderer();
        }

        endCommand();
    }

    public void endCommand() {
        updateMenu();
        commandArea.setActive(true);
        commandQueue.setActive(true);
    }

    public boolean menuActive() {
        return true;
    }

    public void commandKeyTyped() {
    }

    public void focusMoved() {
    }

    // --------------------------------------------------------------------
    // Implements MousePickListener
    // --------------------------------------------------------------------
    public void receiveMousePick(PickResult pickResult) {
    }

    // --------------------------------------------------------------------
    // Menu
    // --------------------------------------------------------------------
    protected abstract void updateMenu();

    // --------------------------------------------------------------------
    // Script
    // --------------------------------------------------------------------
    protected void loadScript(File file) {
        controller.stopRenderer();
        try {
            BufferedReader reader = new BufferedReader(new FileReader(file));
            try {
                String s = null;
                while ((s = reader.readLine()) != null) {
                    String line = s.trim();
                    if (s.startsWith("#")) {
                        continue;
                    }
                    executeSingleCommand(line);
                }
            } finally {
                try {
                    reader.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        } catch (IOException e) {
            showWarning(e);
        } finally {
            controller.startRenderer();
        }
    }

    // --------------------------------------------------------------------
    // Save
    // --------------------------------------------------------------------
    protected void savePng(File file) {
        try {
            BufferedImage image = controller.getImage();
            ImageFileWriter.writePNG(file, image);
        } catch (RecoverableException e) {
            showWarning(e);
        }
    }

    protected void saveJpeg(File file) {
        try {
            BufferedImage image = controller.getImage();
            ImageFileWriter.writeJPEG(file, image, 0.75f);
        } catch (RecoverableException e) {
            showWarning(e);
        }
    }

    // --------------------------------------------------------------------
    // Exception
    // --------------------------------------------------------------------
    protected void showWarning(Exception e) {
        JOptionPane.showMessageDialog(this, e.getMessage(), "Warning",
                JOptionPane.WARNING_MESSAGE);
    }

    protected void showErrorAndExit(Exception e) {
        JOptionPane.showMessageDialog(this, e.getMessage(), "Error",
                JOptionPane.ERROR_MESSAGE);
        exit();
    }

    // --------------------------------------------------------------------
    // Display
    // --------------------------------------------------------------------
    protected void displayProteinWireframe() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("protein", context, false));
        controller.wireframe(0);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayProteinBackbone() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("protein", context, false));
        controller.wireframeOff();
        controller.spacefillOff();
        controller.backbone(0.3f);
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayProteinStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("protein", context, false));
        controller.wireframe(0.2f);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayProteinSpacefill() {
        controller.stopRenderer();
        controller.select(parser.parse("protein", context, false));
        controller.wireframeOff();
        controller.spacefillVdW();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayProteinBallStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("protein", context, false));
        controller.wireframe(0.2f);
        controller.spacefill(0.5f);
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayProteinRibbon() {
        controller.stopRenderer();
        controller.select(parser.parse("protein", context, false));
        controller.wireframeOff();
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbon();
        controller.startRenderer();
    }

    protected void displayProteinCartoon() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("protein", context, false));
        controller.wireframeOff();
        controller.spacefillOff();
        controller.backboneOff();
        controller.cartoon();
        controller.startRenderer();
    }

    protected void displayLigandWireframe() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("LIG", context, false));
        controller.wireframe(0);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("LIG", context, false));
        controller.wireframe(0.2f);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandSpacefill() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("LIG", context, false));
        controller.wireframeOff();
        controller.spacefillVdW();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandBallStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse("LIG", context, false));
        controller.wireframe(0.2f);
        controller.spacefill(0.5f);
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandNeighborWireframe() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse(
                "within(5.0,LIG) & protein & sidechain", context, false));
        controller.wireframe(0);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandNeighborStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse(
                "within(5.0,LIG) & protein & sidechain", context, false));
        controller.wireframe(0.2f);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandNeighborSpacefill() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse(
                "within(5.0,LIG) & protein & sidechain", context, false));
        controller.wireframeOff();
        controller.spacefillVdW();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandNeighborBallStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse(
                "within(5.0,LIG) & protein & sidechain", context, false));
        controller.wireframe(0.2f);
        controller.spacefill(0.5f);
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandNeighborOff() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_AND);
        controller.select(parser.parse(
                "within(5.0,LIG) & protein & sidechain", context, false));
        controller.wireframeOff();
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displayLigandHbondDashedLine() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_OR);
        controller.select(parser.parse("LIG", context, false));
        controller.hBond(0);
        controller.startRenderer();
    }

    protected void displayLigandHbondStick() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_OR);
        controller.select(parser.parse("LIG", context, false));
        controller.hBond(0.2f);
        controller.startRenderer();
    }

    protected void displayLigandHbondOff() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_OR);
        controller.select(parser.parse("LIG", context, false));
        controller.hBondOff();
        controller.startRenderer();
    }

    protected void displaySolventWireframe() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_OR);
        controller.select(parser.parse("SOL", context, false));
        controller.wireframe(0);
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    protected void displaySolventOff() {
        controller.stopRenderer();
        controller.getSelectionStatus().setBondMode(SelectionStatus.BOND_OR);
        controller.select(parser.parse("SOL", context, false));
        controller.wireframeOff();
        controller.spacefillOff();
        controller.backboneOff();
        controller.ribbonOff();
        controller.startRenderer();
    }

    // --------------------------------------------------------------------
    // DSSP
    // --------------------------------------------------------------------
    protected void processDssp(PlotDataXpm data) {
        float scale = data.getXaxisScale();
        int nf = moleculeFile.numMolecules();
        int m = 0;
        for (int ix = 0; ix < data.getNumX(); ix++) {
            float time = data.getX(ix) * scale;
            while (m < nf) {
                MoleculeTraj molecule = moleculeFile.getMoleculeTraj(m);
                float mt = molecule.getTime();
                if (mt < time) {
                    clearStructure(molecule);
                    m++;
                } else if (mt == time) {
                    assignStructure(molecule, data, ix);
                    m++;
                    break;
                } else {
                    break;
                }
            }
        }
        while (m < nf) {
            MoleculeTraj molecule = moleculeFile.getMoleculeTraj(m);
            clearStructure(molecule);
            m++;
        }
        controller.select(parser.parse("protein", context, false));
        ColorExpr colorExpr = new ColorExpr();
        colorExpr.setTarget(ColorExpr.TARGET_RESIDUE);
        colorExpr.setScheme(ColorExpr.SCHEME_DSSP);
        controller.setColor(colorExpr);
        controller.cartoon();
        controller.refresh();
    }

    private void assignStructure(MoleculeTraj molecule, PlotDataXpm data,
            int ix) {
        clearStructure(molecule);
        if (data.getNumY() == 0) {
            return;
        }
        int iy = 0;
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = (Chain) molecule.getChain(ch);
            for (int re = 0; re < chain.numResidues(); re++) {
                Residue residue = (Residue) chain.getResidue(re);
                assignStructure(residue, data.getValue(ix, iy++));
                if (iy == data.getNumY()) {
                    return;
                }
            }
        }
    }

    private void assignStructure(Residue residue, PointValue value) {
        char elem = value.getElement();
        if (elem == '~') {
            residue.setStructure(Residue.STRUCT_NONE);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_COIL);
        } else if (elem == 'E') {
            residue.setStructure(Residue.STRUCT_SHEET);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_B_SHEET);
        } else if (elem == 'B') {
            residue.setStructure(Residue.STRUCT_NONE);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_B_BRIDGE);
        } else if (elem == 'S') {
            residue.setStructure(Residue.STRUCT_TURN);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_BEND);
        } else if (elem == 'T') {
            residue.setStructure(Residue.STRUCT_TURN);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_TURN);
        } else if (elem == 'H') {
            residue.setStructure(Residue.STRUCT_HELIX);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_A_HELIX);
        } else if (elem == 'I') {
            residue.setStructure(Residue.STRUCT_HELIX);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_5_HELIX);
        } else if (elem == 'G') {
            residue.setStructure(Residue.STRUCT_HELIX);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_3_HELIX);
        } else {
            residue.setStructure(Residue.STRUCT_NONE);
            residue.setStructureGromacs(Residue.STRUCT_GROMACS_NONE);
        }
        ColorManager colorManager = controller.getColorManager();
        residue.setColorIndexGromacs(colorManager.getIndex(value.getColor()));
    }

    private void clearStructure(MoleculeTraj molecule) {
        molecule.setStructure(Molecule.STRUCT_GROMACS);
        molecule.setRefreshResidue(true);
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = (Chain) molecule.getChain(ch);
            for (int re = 0; re < chain.numResidues(); re++) {
                Residue residue = (Residue) chain.getResidue(re);
                residue.setStructure(Residue.STRUCT_NONE);
                residue.setStructureGromacs(Residue.STRUCT_GROMACS_NONE);
                residue.setColorIndexGromacs(-1);
            }
        }        
    }

    // --------------------------------------------------------------------
    // H-bond
    // --------------------------------------------------------------------
    protected void processHbond(File indexFile, PlotDataXpm data) {
        int[][] hbonds = readHbondIndex(indexFile);
        if (hbonds.length != data.getNumY()) {
            throw new CustomException(
                    "different number of H-bonds between index file and map file.");
        }
        Map<Integer, Atom> atomMap = new HashMap<Integer, Atom>();
        Map<Integer, Residue> resdiueMap = new HashMap<Integer, Residue>();
        createMap(atomMap, resdiueMap);
        float scale = data.getXaxisScale();
        int nf = moleculeFile.numMolecules();
        int m = 0;
        for (int ix = 0; ix < data.getNumX(); ix++) {
            float time = data.getX(ix) * scale;
            while (m < nf) {
                MoleculeTraj molecule = moleculeFile.getMoleculeTraj(m);
                float mt = molecule.getTime();
                if (mt < time) {
                    clearHbond(molecule);
                    m++;
                } else if (mt == time) {
                    createHbond(molecule, atomMap, resdiueMap, hbonds, data,
                            ix);
                    m++;
                    break;
                } else {
                    break;
                }
            }
        }
        while (m < nf) {
            MoleculeTraj molecule = moleculeFile.getMoleculeTraj(m);
            clearHbond(molecule);
            m++;
        }
        controller.select(true);
        controller.hBond(0);
    }

    private int[][] readHbondIndex(File indexFile) {
        List<String> lines = new ArrayList<String>();
        try {
            BufferedReader reader
                    = new BufferedReader(new FileReader(indexFile));
            try {
                boolean target = false;
                String s = null;
                while ((s = reader.readLine()) != null) {
                    if (target) {
                        lines.add(s.trim());
                    }
                    if (s.startsWith("[ hbonds_")) {
                        target = true;
                    } else if (s.startsWith("[")) {
                        target = false;
                    }
                }
            } finally {
                try {
                    reader.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }

        } catch (IOException e) {
            throw new CustomException(e);
        }
        int[][] hbonds = new int[lines.size()][2];
        for (int i = 0; i < lines.size(); i++) {
            String[] items = lines.get(i).split("\\s+");
            hbonds[i][0] = Integer.parseInt(items[0]);
            hbonds[i][1] = Integer.parseInt(items[2]);
        }
        return hbonds;
    }

    private void createMap(Map<Integer, Atom> atomMap,
            Map<Integer, Residue> residueMap) {
        MoleculeTraj molecule = moleculeFile.getMoleculeTraj(0);
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = (Chain) molecule.getChain(ch);
            for (int re = 0; re < chain.numResidues(); re++) {
                Residue residue = (Residue) chain.getResidue(re);
                for (int at = 0; at < residue.numAtoms(); at++) {
                    Atom atom = (Atom) residue.getAtom(at);
                    if (!atomMap.containsKey(atom.getId())) {
                        atomMap.put(atom.getId(), atom);
                        residueMap.put(atom.getId(), residue);
                    }
                }
            }
        }
    }

    private void clearHbond(MoleculeTraj molecule) {
        molecule.getHBonds().clear();
        molecule.setBuiltHBond(true);
        for (int ch = 0; ch < molecule.numChains(); ch++) {
            Chain chain = (Chain) molecule.getChain(ch);
            for (int re = 0; re < chain.numResidues(); re++) {
                Residue residue = (Residue) chain.getResidue(re);
                residue.setHBondIndex(-1);
            }
        }
    }

    private void createHbond(MoleculeTraj molecule, Map<Integer, Atom> atomMap,
            Map<Integer, Residue> residueMap, int[][] hbonds, PlotDataXpm data,
            int ix) {
        clearHbond(molecule);
        int iy = 0;
        Residue residuePrev = null;
        for (int[] hbondPair : hbonds) {
            if (data.getValue(ix, iy++).getElement() != 'o') {
                continue;
            }
            Atom atom1 = atomMap.get(hbondPair[0]);
            Atom atom2 = atomMap.get(hbondPair[1]);
            if (atom1 == null || atom2 == null) {
                throw new CustomException(
                        "inconsistent atom ID at H-bond: "
                                + hbondPair[0] + " - " + hbondPair[1]);
            }
            Residue residue1 = residueMap.get(hbondPair[0]);
            Residue residue2 = residueMap.get(hbondPair[1]);
            HBond hbond = new HBond();
            hbond.setSrcCA(residue1.getProteinCA());
            hbond.setDstCA(residue2.getProteinCA());
            hbond.setSrcAtom(atom1);
            hbond.setDstAtom(atom2);
            if (residuePrev == null || residuePrev != residue1) {
                residue1.setHBondIndex(molecule.numHBonds());
            }
            molecule.addHBond(hbond);
            residuePrev = residue1;
        }
    }
}